# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import logging
import numpy as np
from scipy.signal import medfilt2d
from osgeo.gdalconst import GA_ReadOnly
from osgeo import gdal
from matplotlib import pylab as plt
from biopal.statistics.utility_statistics import (
    build_filtering_matrix,
    Covariance2D2Correlation2D,
    SKPD_processing,
    main_correlation_estimation_SR,
    covariance_matrix_vec2mat,
)


def UpperThresholdForestHeight(power_cube, opt_str):
    """It finds the last drop of the power spectrum from opt_str.thr times the
     peak of the spectrum. The corresponding elevation is returned.
    
     INPUT
          power_cube: [Nr x Nc x Nz] positive real valued cube of
                      backscattered power
          opt_str.
                  thr: threshold for the spectrum drop (between 0 and 1)
                  z: [Nz x 1] elevation axis
     OUTPUT
           out_str.
                   z: [Nr x Nc] canopy elevation map
    
     DEPENDENCIES
                 FindStep3"""
    Nr, Na, Nz = power_cube.shape
    power_peak = np.max(power_cube, axis=2)
    greater_power_mask = power_cube > power_peak.reshape((Nr, Na, 1)) * opt_str.thr

    last_drop_linear_indices = FindStep3(greater_power_mask, "last")

    class out_str:
        pass

    out_str.z = opt_str.z[last_drop_linear_indices]
    out_str.peak = power_peak
    return out_str


def FindStep3(I, varargin):
    """This routin finds the first or the last step in a logical 3D along the
     third direction. The first step is a rising one whereas the last is a
     drop.
    
     INPUT
          I: [Nr x Nc x N] logical array to be explored
          varargin{1}: (optional. Default: 'first') either 'first' or 'last'
    
     OUTPUT
           linear_indices: [Nr x Nc] matrix of linear indices where the step
                           has been detected"""
    Nr, Na, Nz = I.shape
    I = I > 0

    if varargin == "first":

        rise_mask = 1 - (np.roll(I, (0, 0, 1), (0, 1, 2))) & I
        rise_mask[:, :, 0] = False
        rise_mask[:, :, -1] = False

        rise_mask_cumsum = np.cumsum(rise_mask, axis=2)

        linear_indices = np.sum(
            ((rise_mask_cumsum == 1) & rise_mask) * np.arange(I.shape[2]).reshape(1, 1, I.shape[2]), axis=2,
        )

    elif varargin == "last":

        drop_mask = 1 - (np.roll(I, (0, 0, -1), (0, 1, 2))) & I
        drop_mask[:, :, 0] = False
        drop_mask[:, :, -1] = False

        drop_mask_cumsum = np.cumsum(drop_mask, axis=2)

        linear_indices = np.sum(
            ((drop_mask_cumsum == drop_mask_cumsum[:, :, -1].reshape((Nr, Na, 1))) & drop_mask)
            * np.arange(I.shape[2]).reshape(1, 1, I.shape[2]),
            axis=2,
        )

    else:

        print("Unrecognized input, aborting.\n")
        linear_indices = []

    return linear_indices


def BiomassForestHeightSKPD(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    carrier_frequency_hz,
    range_bandwidth_hz,
    kz_stack,
    vertical_vector,
    proc_conf,
):

    power_threshold = proc_conf.power_threshold

    # data_stack is a dictionary of two nested dictionaries composed as:
    # data_stack[ acquisition_name ][ polarization ]

    num_acq = len(data_stack)
    acq_names = list(data_stack.keys())
    first_acq_dict = data_stack[acq_names[0]]
    pol_names = list(first_acq_dict.keys())
    num_pols = len(pol_names)
    Nrg, Naz = first_acq_dict[pol_names[0]].shape
    Nz = np.size(vertical_vector)

    # Covariance estimation
    (MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_correlation_estimation_SR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_slant_rg,
        pixel_spacing_az,
        incidence_angle_rad,
        carrier_frequency_hz,
        range_bandwidth_hz,
    )

    Nrg_subs = rg_vec_subs.size
    Naz_subs = az_vec_subs.size

    # Initialization of the SKPD routine
    class SKPD_kernel_opt_str:
        pass

    SKPD_kernel_opt_str.num_acq = num_acq
    SKPD_kernel_opt_str.num_pols = num_pols
    SKPD_kernel_opt_str.Nsubspaces = 2
    SKPD_kernel_opt_str.Nparam = 2
    SKPD_kernel_opt_str.error = np.zeros((Nrg_subs, Naz_subs, 4, 2))

    # Single polarimetric channel selector
    wpol = (np.kron(np.eye(num_pols), np.ones((1, num_acq)))) > 0

    tomo_cube = np.zeros((Nrg_subs, Naz_subs, Nz))
    Nrg_subs_string = str(Nrg_subs)
    for rg_sub_idx in np.arange(Nrg_subs):

        logging.info("   Heigth step " + str(rg_sub_idx + 1) + " of " + Nrg_subs_string)
        for az_sub_idx in np.arange(Naz_subs):

            # Spectra estimation initialization
            class spectra:
                pass

            spectra.temp = np.zeros((Nz, 4))
            current_MPMB_correlation = MPMB_correlation[:, :, rg_sub_idx, az_sub_idx]

            # SKPD processing
            SKPD_kernel_out_str = SKPD_processing(current_MPMB_correlation, SKPD_kernel_opt_str)

            if np.any(SKPD_kernel_out_str.error):
                # Calibrating with respect to the linked phases of the
                # best polarimetric channel
                Rcoh_thin = Covariance2D2Correlation2D(
                    np.reshape(
                        current_MPMB_correlation[wpol[0, :][:, np.newaxis] * wpol[0, :][np.newaxis, :]],
                        (num_acq, num_acq),
                    )
                )
                Rcoh_fat = Covariance2D2Correlation2D(
                    np.reshape(
                        current_MPMB_correlation[wpol[1, :][:, np.newaxis] * wpol[1, :][np.newaxis, :]],
                        (num_acq, num_acq),
                    )
                )
                Rincoh_thin = Covariance2D2Correlation2D(
                    np.reshape(
                        current_MPMB_correlation[wpol[2, :][:, np.newaxis] * wpol[2, :][np.newaxis, :]],
                        (num_acq, num_acq),
                    )
                )
                Rincoh_fat = np.random.randn(num_acq, num_acq) + 1j * np.random.randn(num_acq, num_acq)

            else:
                # Scattering mechanisms
                Rcoh_thin = Covariance2D2Correlation2D(SKPD_kernel_out_str.Rcoh_thin)
                Rcoh_fat = Covariance2D2Correlation2D(SKPD_kernel_out_str.Rcoh_fat)
                Rincoh_thin = Covariance2D2Correlation2D(SKPD_kernel_out_str.Rincoh_thin)
                Rincoh_fat = Covariance2D2Correlation2D(SKPD_kernel_out_str.Rincoh_fat)

            # Steering matrix
            current_kz = np.zeros((num_acq, 1))
            for b_idx, stack_curr in enumerate(kz_stack.values()):
                current_kz[b_idx] = stack_curr[rg_vec_subs[rg_sub_idx], az_vec_subs[az_sub_idx]]

            A = np.exp(1j * current_kz * vertical_vector) / num_acq

            # Spectra estimation
            for m in np.arange(4):
                currR = (m == 0) * Rcoh_thin + (m == 1) * Rcoh_fat + (m == 2) * Rincoh_thin + (m == 3) * Rincoh_fat
                if proc_conf.enable_super_resolution:
                    # Capon
                    currR = currR + proc_conf.regularization_noise_factor * np.eye(currR.shape[0])
                    spectra.temp[:, m] = 1 / np.diag(np.abs(A.conj().transpose() @ np.linalg.inv(currR) @ A))
                else:
                    spectra.temp[:, m] = np.diag(np.abs(A.conj().transpose() @ currR @ A))

            # Volume mechanism recognized thanks to its higher elevation
            max_index = np.argmax(spectra.temp, axis=0)
            max_m = np.argmax(max_index)
            tomo_cube[rg_sub_idx, az_sub_idx, :] = spectra.temp[:, max_m]

    # Estimating canopy elevation by looking at the decay
    class opt_str:
        pass

    opt_str.z = vertical_vector
    opt_str.thr = power_threshold  # Power decay threshold With respect to the peak value
    out_str = UpperThresholdForestHeight(tomo_cube, opt_str)
    canopy_height = out_str.z
    power_peak = out_str.peak
    canopy_height = medfilt2d(canopy_height.astype("float64"), kernel_size=5)

    return canopy_height, power_peak, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def tomo_cube_generator(z, I, model_sign, kz_stack):
    # Beamforming (Fourier spectrum estimation)
    Nz = np.size(z)
    Nr, Na, N = I.shape
    # Initialization
    tomo_cube = np.zeros((Nr, Na, Nz), dtype=np.complex128)

    print("        Beamforming...")
    for z_ind in np.arange(Nz):
        steering_array = np.exp(-1j * model_sign * kz_stack * z[z_ind]) / np.sqrt(N)
        tomo_cube[:, :, z_ind] = np.sum(I * steering_array, axis=2)
    print("        done.")
    return tomo_cube


def tomo_power_computation_OLD(savPath, win_x, win_y, pol_string, save_flag=False):
    loadPath = os.path.abspath(savPath)
    loadlist = os.listdir(loadPath)
    for ind in loadlist:
        if "x_ax" in ind:
            x_ax = np.load(os.path.join(loadPath, ind))
            Na = np.size(x_ax)
            da = x_ax[1] - x_ax[0]
        if "y_ax" in ind:
            y_ax = np.load(os.path.join(loadPath, ind))
            Nr = np.size(y_ax)
            dr = y_ax[1] - y_ax[0]
        if "z_ax" in ind:
            z_ax = np.load(os.path.join(loadPath, ind))
            Nz = np.size(z_ax)

    # Range filter matrix
    class filtering_matrix_opt_str:
        pass

    NW = np.round(win_y / dr).astype("int64")
    filtering_matrix_opt_str.Nin = Nr
    filtering_matrix_opt_str.NWmono = np.floor((NW - 1) / 2).astype("int64")
    filtering_matrix_opt_str.subs_F = filtering_matrix_opt_str.NWmono
    Fr, rg_subs_ind, Rnorm = build_filtering_matrix(filtering_matrix_opt_str)
    Nrg_subs = rg_subs_ind.size
    Fr_normalized = Rnorm @ Fr

    # Azimuth filter matrix
    NW = np.round(win_x / da).astype("int64")
    filtering_matrix_opt_str.Nin = Na
    filtering_matrix_opt_str.NWmono = np.floor((NW - 1) / 2).astype("int64")
    filtering_matrix_opt_str.subs_F = filtering_matrix_opt_str.NWmono
    Fa, az_subs_ind, Anorm = build_filtering_matrix(filtering_matrix_opt_str)
    Naz_subs = az_subs_ind.size
    Fa_normalized = (Anorm @ Fa).T

    # Initialization
    power_cube = np.zeros((Nrg_subs, Naz_subs, Nz))
    n = 0
    ## Backscattered power estimation
    print("Power estimation...")
    counter = 0
    for fileName in loadlist:
        if "_tomo_" in fileName:
            fileName_split = fileName.split("_")
            current_pol_string = fileName_split[-3]
            slice_idx = int(fileName_split[-1].split(".")[0])

            if pol_string == current_pol_string:

                slice_loaded = np.load(os.path.join(loadPath, fileName))

                counter = counter + 1
                print("    adding slice " + str(counter) + " of " + str(Nz))
                power_cube[:, :, slice_idx] = Fr_normalized @ np.abs(slice_loaded) ** 2 @ Fa_normalized
                n += 1
                print("    {:d}".format(int(n / Nz * 100)) + "%")

    x_ax = x_ax[az_subs_ind]
    y_ax = y_ax[rg_subs_ind]

    if save_flag:
        np.save(os.path.join(savPath, "power_cube"), power_cube)
        np.save(os.path.join(savPath, "x_ax_power_cube"), x_ax)
        np.save(os.path.join(savPath, "x_ax_power_cube"), y_ax)
        np.save(os.path.join(savPath, "x_ax_power_cube"), z_ax)

    return x_ax, y_ax, z_ax, power_cube


def tomo_power_computation(cube_sllices_folder, equi7_zone_name, equi7_tile_name, win_x, win_y, save_path=""):

    start_a = 4460
    start_r = 3880
    Na = 140
    Nr = 160

    slices_folders = os.listdir(cube_sllices_folder)
    Nz = len(slices_folders)

    ## Backscattered power estimation
    print("Power estimation...")
    for slice_idx, slice_folder in enumerate(slices_folders):

        cube_slice_fname = os.path.join(
            cube_sllices_folder,
            slice_folder,
            equi7_zone_name,
            equi7_tile_name,
            "tomo_cube_slice_" + equi7_zone_name[6:] + "_" + equi7_tile_name + ".tif",
        )

        data_driver = gdal.Open(cube_slice_fname, GA_ReadOnly)
        for cov_idx in np.arange(6):
            if cov_idx == 0:
                cov_vec_curr = data_driver.GetRasterBand(int(cov_idx + 1)).ReadAsArray(start_r, start_a, Nr, Na).T
                plt.figure()
                plt.imshow(cov_vec_curr, cmap="jet")
                plt.pause(1)

                Nr = cov_vec_curr.shape[0]
                Na = cov_vec_curr.shape[1]
                cov_vec = np.zeros((6, Nr, Na))
                cov_vec[cov_idx, :, :] = cov_vec_curr
            else:
                cov_vec[cov_idx, :, :] = (
                    data_driver.GetRasterBand(int(cov_idx + 1)).ReadAsArray(start_r, start_a, Nr, Na).T
                )

        print("reshape input cov...")
        cov_mat = covariance_matrix_vec2mat(cov_vec)
        print("done")

        if slice_idx == 0:
            # Initialization

            geotransform_equi7 = data_driver.GetGeoTransform()
            dr = np.abs(geotransform_equi7[1])
            da = np.abs(geotransform_equi7[5])

            # Range filter matrix
            class filtering_matrix_opt_str:
                pass

            NW = np.round(win_y / dr).astype("int64")
            filtering_matrix_opt_str.Nin = Nr
            filtering_matrix_opt_str.NWmono = np.floor((NW - 1) / 2).astype("int64")
            filtering_matrix_opt_str.subs_F = filtering_matrix_opt_str.NWmono
            Fr, rg_subs_ind, Rnorm = build_filtering_matrix(filtering_matrix_opt_str)
            Nrg_subs = rg_subs_ind.size
            Fr_normalized = Rnorm @ Fr

            # Azimuth filter matrix
            NW = np.round(win_x / da).astype("int64")
            filtering_matrix_opt_str.Nin = Na
            filtering_matrix_opt_str.NWmono = np.floor((NW - 1) / 2).astype("int64")
            filtering_matrix_opt_str.subs_F = filtering_matrix_opt_str.NWmono
            Fa, az_subs_ind, Anorm = build_filtering_matrix(filtering_matrix_opt_str)
            Naz_subs = az_subs_ind.size
            Fa_normalized = (Anorm @ Fa).T

            power_cube = np.zeros((Nrg_subs, Naz_subs, Nz))

        print("    adding " + slice_folder)
        power_cube[:, :, slice_idx] = Fr_normalized @ np.abs(cov_mat) ** 2 @ Fa_normalized

    if save_path:
        np.save(os.path.join(save_path, "power_cube"), power_cube)

    return power_cube


def tomo_plot(
    cube_slices_folder,
    pol_name,
    east_slice_index,
    north_slice_index,
    z_slice_index,
    equi7_zone_name,
    equi7_tile_name,
    plot_flag,
    pixel_extent="",
    save_power_cube_path="",
):

    if pol_name == "hh":
        pol_idx = 0
    elif pol_name == "hv":
        pol_idx = 3
    elif pol_name == "vv":
        pol_idx = 5

    z_file = os.path.join(cube_slices_folder, "vertical_vector.txt")
    if not os.path.exists(z_file):
        raise ValueError(
            'A file named "vertical_vector.txt" should be present in the slices folder ' + cube_slices_folder
        )
    z_file_id = open(z_file, "r")
    slices_folders = os.listdir(cube_slices_folder)
    num_z = len(slices_folders)
    z_ax = np.zeros(num_z)
    for slice_idx, cube_slice_curr in enumerate(slices_folders):
        if not "tomo_cube_slice_" in cube_slice_curr:
            continue

        print("Reading slice {} of {}".format(slice_idx + 1, num_z))
        line_curr = z_file_id.readline()
        idx_altitude = line_curr.find("altitude:")

        z_ax[slice_idx] = line_curr[idx_altitude + 10 : -5]

        cube_slice_fname = os.path.join(
            cube_slices_folder,
            cube_slice_curr,
            equi7_zone_name,
            equi7_tile_name,
            "tomo_cube_slice_" + equi7_zone_name[6:] + "_" + equi7_tile_name + ".tif",
        )
        print("cube_slice_fname: ", cube_slice_fname)
        data_driver = gdal.Open(cube_slice_fname, GA_ReadOnly)
        if slice_idx == 0:
            TEMP = data_driver.GetRasterBand(int(pol_idx + 1)).ReadAsArray()

            tomo_cube = np.zeros((num_z, TEMP.shape[0], TEMP.shape[1]))

            geotransform_equi7 = data_driver.GetGeoTransform()
            east_len = TEMP.shape[0]
            north_len = TEMP.shape[1]

            east_0 = geotransform_equi7[0]
            east_end = east_0 + geotransform_equi7[1] * east_len
            north_0 = geotransform_equi7[3]
            north_end = north_0 + geotransform_equi7[5] * north_len

            north_ax = np.linspace(0, north_len, num=north_len) * geotransform_equi7[5] + north_0
            east_ax = np.linspace(0, east_len, num=east_len) * geotransform_equi7[1] + east_0

        tomo_cube[slice_idx, :, :] = np.abs(data_driver.GetRasterBand(int(pol_idx + 1)).ReadAsArray())

    if 1:
        if plot_flag == "db":

            def fun_plot(temp_plot):
                return np.log10(temp_plot)

        else:

            def fun_plot(temp_plot):
                return temp_plot

        if pixel_extent:
            east_0 = east_ax[pixel_extent[0]] / 1000
            east_end = east_ax[pixel_extent[1]] / 1000
            north_0 = north_ax[pixel_extent[2]] / 1000
            north_end = north_ax[pixel_extent[3]] / 1000

        plt.figure()
        tomo_plt = np.squeeze(tomo_cube[z_slice_index, :, :])
        if pixel_extent:
            tomo_plt = tomo_plt[pixel_extent[0] : pixel_extent[1], pixel_extent[2] : pixel_extent[3]]
        ax = plt.subplot(3, 1, 1)

        plt.imshow(
            fun_plot(tomo_plt.T),
            interpolation="none",
            origin="upper",
            extent=[east_0, east_end, north_end, north_0],
            cmap="jet",
        )
        plt.title("z = {:3.0f} [m]".format(z_ax[z_slice_index]))
        plt.xlabel("east [Km]")
        plt.ylabel("north [Km]")
        ax.set_aspect("auto")
        tomo_plt = np.squeeze(tomo_cube[:, :, north_slice_index])
        if pixel_extent:
            tomo_plt = tomo_plt[:, pixel_extent[0] : pixel_extent[1]]
        ax = plt.subplot(3, 1, 2)
        plt.imshow(
            fun_plot(tomo_plt),
            interpolation="none",
            origin="lower",
            extent=[east_0, east_end, z_ax[0], z_ax[-1]],
            cmap="jet",
        )
        plt.title("north = {:3.0f} [Km]".format(north_ax[north_slice_index] / 1000))
        plt.xlabel("east [Km]")
        plt.ylabel("z [m]")
        ax.set_aspect("auto")
        tomo_plt = np.squeeze(tomo_cube[:, east_slice_index, :])
        if pixel_extent:
            tomo_plt = tomo_plt[:, pixel_extent[2] : pixel_extent[3]]
        ax = plt.subplot(3, 1, 3)
        plt.imshow(
            fun_plot(tomo_plt),
            interpolation="none",
            origin="lower",
            extent=[north_0, north_end, z_ax[0], z_ax[-1]],
            cmap="jet",
        )
        plt.title("east = {:3.0f} [Km]".format(east_ax[east_slice_index] / 1000))
        plt.xlabel("north [Km]")
        plt.ylabel("z [m]")
        ax.set_aspect("auto")
        plt.subplots_adjust(hspace=0.4, wspace=0.2)

        plt.figure()
        tomo_plt = np.squeeze(tomo_cube[:, east_slice_index, north_slice_index])
        plt.plot(fun_plot(tomo_plt))
        plt.title(
            "north = {} [Km], east = {} [Km]".format(
                north_ax[north_slice_index] / 1000, east_ax[east_slice_index] / 1000
            )
        )
        plt.xlabel("z [m]")
        if plot_flag == "db":
            plt.ylabel(plot_flag)
        ax.set_aspect("auto")

        if save_power_cube_path:
            np.save(os.path.join(save_power_cube_path, "tomo_cube.npy"), tomo_cube)



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
General SAR attitude module
---------------------------
"""

import os
from typing import Union
from enum import Enum
import numpy as np

from arepytools import _utils
from arepytools.math import axis as are_ax
from arepytools.geometry.generalsarorbit import GeneralSarOrbit, create_general_sar_orbit
from arepytools.io.metadata import AttitudeInfo, StateVectors
from arepytools.geometry._interpolator import GeometryInterpolator


class ReferenceFrame(Enum):
    """Attitude reference frame"""
    geocentric = 'GEOCENTRIC'
    geodetic = 'GEODETIC'
    zero_doppler = 'ZERODOPPLER'


class Angles(Enum):
    """Attitude angles (Yaw, Pitch, Roll)"""
    Y = 0
    P = 1
    R = 2


class RotationOrder(Enum):
    """Yaw / Pitch / Roll rotation order"""
    ypr = 'YPR'
    yrp = 'YRP'
    pry = 'PRY'
    pyr = 'PYR'
    ryp = 'RYP'
    rpy = 'RPY'


class EnumerateRot:
    """Iterate on angles of a given a rotation order.

    Example:

    .. code-block:: python

        for angleId in EnumerateRot(RotationOrder('PRY')):
            print("{}, corresponding id: {}".format(Angles(angleId), angleId))
    """

    def __init__(self, rotation_order: RotationOrder):
        """Initialize the object to iterate on the given rotation order"""
        self._rotation_order = rotation_order.value
        self._current = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._current < 3:
            self._current += 1
            return Angles[self._rotation_order[self._current - 1]].value
        else:
            raise StopIteration()


class GeneralSarAttitude:
    """
    GeneralSarAttitude class
    """

    _MINIMUM_NUMBER_OF_DATA_POINTS = GeometryInterpolator.get_min_num_of_data_points()

    @classmethod
    def get_minimum_number_of_data_points(cls):
        """Return the required minimum number of attitude data points

        :return: required minimum number of attitude data points
        """
        return cls._MINIMUM_NUMBER_OF_DATA_POINTS

    @property
    def t0(self):
        """Origin of the time axis"""
        return self._time_axis.start

    @property
    def dt(self):
        """Time axis step (if applicable)"""
        if isinstance(self._time_axis, are_ax.RegularAxis):
            return self._time_axis.step
        raise RuntimeError("Time step is not available for attitudes constructed with non-regular time axis")

    @property
    def n(self):
        """Number of attitude data points"""
        return self._time_axis.size

    @property
    def ypr_angles(self):
        """YPR angles as a 3xN numpy array"""
        return self._ypr_angles

    @property
    def rotation_order(self):
        """Yaw/Pitch/Roll rotation order"""
        return self._rotation_order

    @property
    def reference_frame(self):
        """Reference frame"""
        return self._reference_frame

    @property
    def time_axis_array(self):
        """Time axis as array of time points"""
        return self._time_axis.get_array()

    @property
    def interpolator(self):
        """Geometry interpolator object"""
        return self._interpolator

    def __init__(self, orbit: GeneralSarOrbit, time_axis: Union[np.ndarray, are_ax.Axis],
                 ypr_angles: np.ndarray, rotation_order: str, reference_frame: str):
        """Initialize the general sar attitude using the provided input parameters

        :param orbit: sensor orbit as a general sar orbit object
        :param time_axis: time axis of length N, as Axis or numpy array of PreciseDateTime objects.
        :param ypr_angles: yaw, pitch and roll angles as 3xN numpy array
        :param rotation_order: rotation order string
        :param reference_frame: reference frame string
        """

        if isinstance(time_axis, np.ndarray):
            time_axis_start = time_axis[0]
            relative_time_axis = (time_axis - time_axis_start).astype(float)
            self._time_axis = are_ax.Axis(relative_time_axis, time_axis_start)
        else:
            self._time_axis = time_axis

        _check_init_input(orbit, self._time_axis, ypr_angles, rotation_order, reference_frame)

        self._orbit = orbit
        self._rotation_order = RotationOrder(rotation_order.upper())
        self._reference_frame = ReferenceFrame(reference_frame.upper())
        self._ypr_angles = ypr_angles

        self._interpolator = GeometryInterpolator(self._time_axis, self._ypr_angles)

    def _interpolate_angles(self, time_points, angles_to_interpolate, interval_indexes):
        """Interpolate the required angle components on the given time points

        :param time_points: 1D numpy array of length N of absolute time points
        :param angles_to_interpolate: C-length vector of indexes of the angle components to interpolate
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: CxN numpy array of interpolated angle values"""
        return self.interpolator.eval(time_points, interval_indexes, angles_to_interpolate)

    def get_yaw(self, time_points, interval_indexes=None):
        """Return the yaw angles at the specified time points

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 1D numpy array of N yaw angles
        """
        return self._interpolate_angles(time_points, [Angles.Y.value], interval_indexes)

    def get_pitch(self, time_points, interval_indexes=None):
        """Return the pitch angles at the specified time points

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 1D numpy array of N pitch angles
        """
        return self._interpolate_angles(time_points, [Angles.P.value], interval_indexes)

    def get_roll(self, time_points, interval_indexes=None):
        """Return the roll angles at the specified time points

        :param time_points: 1D numpy array of length N of absolute time points
        :param interval_indexes: intervals of the time axis where the given time points are expected [optional]

        :return: 1D numpy array of N roll angles
        """
        return self._interpolate_angles(time_points, [Angles.R.value], interval_indexes)

    def __repr__(self):
        axis_str = str(self._time_axis)
        ypr_str = str(self._ypr_angles)
        gso_str = str(self._orbit)

        axis_portion = "Attitude defined on azimuth axis: " + os.linesep + axis_str + os.linesep
        state_vectors_portion = "Yaw Pitch Roll matrix: " + os.linesep + ypr_str + os.linesep
        rotation_order = "Rotation order: {}".format(self.rotation_order.name.upper()) + os.linesep
        reference_frame = "Reference frame: {}".format(self.reference_frame.name.upper()) + os.linesep
        gso_portion = "Attitude info base on orbit:" + os.linesep + gso_str + os.linesep
        return axis_portion + state_vectors_portion + rotation_order + reference_frame + gso_portion


def _check_init_input(gso, azimuth_axis, ypr_matrix, rotation_order, reference_frame):
    _utils.check_type(gso, GeneralSarOrbit, 'gso')
    _utils.check_type(azimuth_axis, are_ax.Axis, 'axis')
    _utils.check_type(ypr_matrix, np.ndarray, 'ypr_matrix')
    _utils.check_ndim_of_numpy_array(ypr_matrix, 2, 'ypr_matrix')
    _utils.check_first_axis_size_of_numpy_array(ypr_matrix, 3, 'ypr_matrix')
    if ypr_matrix.shape[1] < GeneralSarAttitude.get_minimum_number_of_data_points():
        raise RuntimeError(
            "Not enough attitude records provided: {} < {}".format(
                ypr_matrix.shape[1], GeneralSarAttitude.get_minimum_number_of_data_points()))
    if ypr_matrix.shape[1] != azimuth_axis.size:
        raise RuntimeError(
            "Time and attitude records matrix shall have same number of elements: {} != {}".format(
                ypr_matrix.shape[1], azimuth_axis.size))
    _utils.check_dtype_of_numpy_array(ypr_matrix, float, 'ypr_matrix')
    _utils.check_type(rotation_order, str, 'rotation_order')
    _utils.check_type(reference_frame, str, 'reference_frame')


def create_general_sar_attitude(state_vectors: StateVectors, attitude_info: AttitudeInfo) -> GeneralSarAttitude:
    """Create general sar attitude object from state vectors and attitude info metadata

    :param state_vectors: state vectors as a StateVectors metadata object
    :param attitude_info: attitude data as a AttitudeInfo metadata object

    :return: the new GeneralSarAttitude object
    """
    gso = create_general_sar_orbit(state_vectors)
    time_axis = are_ax.RegularAxis((0, attitude_info.time_step, attitude_info.attitude_records_number),
                                   attitude_info.reference_time)
    ypr_matrix = np.vstack((attitude_info.yaw_vector, attitude_info.pitch_vector, attitude_info.roll_vector))
    gsa = GeneralSarAttitude(gso, time_axis, ypr_matrix, attitude_info.rotation_order.value,
                             attitude_info.reference_frame.value)
    return gsa



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
pyxb to aresys conversion module
--------------------------------
"""
import logging
import pyxb
import numpy as np
from arepytools import constants
from . import pyxb_metadata_types, pyxb_metadata
from .. import metadata
from ...timing.precisedatetime import PreciseDateTime


def convert_sampling_constants_to_aresys(
        i_sampling_constants: pyxb_metadata_types.SamplingConstantsType) -> metadata.SamplingConstants:
    o_sampling_constants = metadata.SamplingConstants(i_sampling_constants.frg_hz.value(),
                                                      i_sampling_constants.Brg_hz.value(),
                                                      i_sampling_constants.faz_hz.value(),
                                                      i_sampling_constants.Baz_hz.value())

    return o_sampling_constants


def convert_pulse_to_aresys(i_pulse: pyxb_metadata_types.PulseType) -> metadata.Pulse:
    o_pulse = metadata.Pulse(i_pulse.PulseLength.value(),
                             i_pulse.Bandwidth.value(),
                             i_pulse.PulseSamplingRate.value(),
                             i_pulse.PulseEnergy.value(),
                             )
    o_pulse.pulse_direction = i_pulse.Direction
    if i_pulse.PulseStartFrequency is not None:
        o_pulse.pulse_start_frequency = i_pulse.PulseStartFrequency.value()
    if i_pulse.PulseStartPhase is not None:
        o_pulse.pulse_start_phase = i_pulse.PulseStartPhase.value()

    return o_pulse


def convert_swath_info_to_aresys(i_swath_info: pyxb_metadata_types.SwathInfoType) -> metadata.SwathInfo:
    o_swath_info = metadata.SwathInfo(i_swath_info.Swath.value(),
                                      i_swath_info.Polarization,
                                      i_swath_info.AcquisitionPRF)

    o_swath_info.swath_acquisition_order = i_swath_info.SwathAcquisitionOrder.value()
    o_swath_info.rank = i_swath_info.Rank.value()
    o_swath_info.range_delay_bias = i_swath_info.RangeDelayBias.value()
    o_swath_info.acquisition_start_time = PreciseDateTime().set_from_utc_string(
        i_swath_info.AcquisitionStartTime.value())
    o_swath_info.azimuth_steering_rate_reference_time = i_swath_info.AzimuthSteeringRateReferenceTime.value()
    o_swath_info.azimuth_steering_rate_pol = (i_swath_info.AzimuthSteeringRatePol.val[0].value(),
                                              i_swath_info.AzimuthSteeringRatePol.val[1].value(),
                                              i_swath_info.AzimuthSteeringRatePol.val[2].value())
    o_swath_info.echoes_per_burst = i_swath_info.EchoesPerBurst
    if i_swath_info.ChannelDelay is not None:
        o_swath_info.channel_delay = i_swath_info.ChannelDelay
    if i_swath_info.RxGain is not None:
        o_swath_info.rx_gain = i_swath_info.RxGain
    return o_swath_info


def convert_rasterinfo_to_aresys(i_rasterinfo: pyxb_metadata_types.RasterInfoType):
    samples = i_rasterinfo.Samples
    samples_start = float(i_rasterinfo.SamplesStart.value())
    samples_start_unit = i_rasterinfo.SamplesStart.unit
    samples_step = float(i_rasterinfo.SamplesStep.value())
    samples_step_unit = i_rasterinfo.SamplesStep.unit
    lines = i_rasterinfo.Lines
    lines_start_unit = i_rasterinfo.LinesStart.unit
    if lines_start_unit.lower() == 'utc':
        lines_start = PreciseDateTime()
        lines_start.set_from_utc_string(i_rasterinfo.LinesStart.value())
    else:
        lines_start = float(i_rasterinfo.LinesStart.value())
    lines_step = float(i_rasterinfo.LinesStep.value())
    lines_step_unit = i_rasterinfo.LinesStep.unit
    celltype = i_rasterinfo.CellType
    filename = i_rasterinfo.FileName
    header_offset_bytes = i_rasterinfo.HeaderOffsetBytes
    row_prefix_bytes = i_rasterinfo.RowPrefixBytes
    byteorder = i_rasterinfo.ByteOrder

    invalid_value = i_rasterinfo.InvalidValue
    if invalid_value is not None:
        invalid_value = complex(invalid_value.RealValue, invalid_value.ImaginaryValue)

    format_type = i_rasterinfo.RasterFormat
    if format_type is not None:
        format_type = str(format_type)

    o_rasterinfo = metadata.RasterInfo(lines=lines, samples=samples, celltype=celltype,
                                       filename=filename,
                                       header_offset_bytes=header_offset_bytes,
                                       row_prefix_bytes=row_prefix_bytes,
                                       byteorder=byteorder, invalid_value=invalid_value,
                                       format_type=format_type)

    o_rasterinfo.set_lines_axis(lines_start, lines_start_unit, lines_step, lines_step_unit)
    o_rasterinfo.set_samples_axis(samples_start, samples_start_unit, samples_step, samples_step_unit)

    return o_rasterinfo


def convert_dataset_info_to_aresys(i_dataset_info: pyxb_metadata_types.DataSetInfoType):
    o_dataset_info = metadata.DataSetInfo(i_dataset_info.AcquisitionMode.value(), i_dataset_info.fc_hz.value())

    o_dataset_info.sensor_name = i_dataset_info.SensorName
    o_dataset_info.description = i_dataset_info.Description.value()
    sense_date = i_dataset_info.SenseDate.value()
    if sense_date == 'NOT_AVAILABLE' or sense_date is None:
        o_dataset_info.sense_date = None
    else:
        o_dataset_info.sense_date = \
            PreciseDateTime().set_from_utc_string(sense_date)
    o_dataset_info.image_type = i_dataset_info.ImageType.value()
    o_dataset_info.projection = i_dataset_info.Projection.value()
    o_dataset_info.acquisition_station = i_dataset_info.AcquisitionStation.value()
    o_dataset_info.processing_center = i_dataset_info.ProcessingCenter.value()
    process_date = i_dataset_info.ProcessingDate.value()
    if process_date == 'NOT_AVAILABLE' or process_date is None:
        o_dataset_info.sense_date = None
    else:
        o_dataset_info.processing_date = \
            PreciseDateTime().set_from_utc_string(process_date)
    o_dataset_info.processing_software = i_dataset_info.ProcessingSoftware.value()
    o_dataset_info.side_looking = i_dataset_info.SideLooking
    o_dataset_info.external_calibration_factor = i_dataset_info.ExternalCalibrationFactor
    o_dataset_info.data_take_id = i_dataset_info.DataTakeID

    return o_dataset_info


def convert_state_vectors_to_aresys(i_state_vectors: pyxb_metadata_types.StateVectorDataType):
    number_of_state_vectors = i_state_vectors.nSV_n.value()

    position_vector = np.zeros((number_of_state_vectors, 3))
    velocity_vector = np.zeros((number_of_state_vectors, 3))
    for ecef_coord_index, ecef_coord in enumerate(i_state_vectors.pSV_m.val):
        position_vector[ecef_coord_index // 3][ecef_coord_index % 3] = \
            ecef_coord.value()
        velocity_vector[ecef_coord_index // 3][ecef_coord_index % 3] = \
            i_state_vectors.vSV_mOs.val[ecef_coord_index].value()

    reference_time = PreciseDateTime()
    reference_time.set_from_utc_string(i_state_vectors.t_ref_Utc)
    delta_time = i_state_vectors.dtSV_s.value()

    o_state_vectors = metadata.StateVectors(position_vector,
                                            velocity_vector,
                                            reference_time,
                                            delta_time)

    if i_state_vectors.OrbitNumber != 'NOT_AVAILABLE':
        o_state_vectors.orbit_number = int(i_state_vectors.OrbitNumber)

    if i_state_vectors.Track != 'NOT_AVAILABLE':
        o_state_vectors.track_number = int(i_state_vectors.Track)

    return o_state_vectors


def convert_attitude_info_to_aresys(i_attitude_info: pyxb_metadata_types.AttitudeInfoType):
    o_attitude_info = metadata.AttitudeInfo(yaw=_pyxb_sequence_to_list(i_attitude_info.yaw_deg),
                                            pitch=_pyxb_sequence_to_list(i_attitude_info.pitch_deg),
                                            roll=_pyxb_sequence_to_list(i_attitude_info.roll_deg),
                                            t0=PreciseDateTime().set_from_utc_string(i_attitude_info.t_ref_Utc),
                                            delta_t=i_attitude_info.dtYPR_s.value(),
                                            ref_frame=i_attitude_info.referenceFrame,
                                            rot_order=i_attitude_info.rotationOrder)

    o_attitude_info.attitude_type = i_attitude_info.AttitudeType

    return o_attitude_info


def _pyxb_sequence_to_list(pyxb_seq):
    if pyxb_seq is None:
        return []
    return [v.value() for v in pyxb_seq.val]


def convert_acquisition_time_line_to_aresys(i_acquisition_time_line: pyxb_metadata_types.AcquisitionTimelineType):
    swl_changes_number = i_acquisition_time_line.Swl_changes_number if i_acquisition_time_line.Swl_changes_number else 0
    o_acquisition_time_line = metadata.AcquisitionTimeLine(
        missing_lines_number_i=i_acquisition_time_line.MissingLines_number.value(),
        missing_lines_azimuth_times_i=_pyxb_sequence_to_list(i_acquisition_time_line.MissingLines_azimuthtimes),
        swst_changes_number_i=i_acquisition_time_line.Swst_changes_number.value(),
        swst_changes_azimuth_times_i=_pyxb_sequence_to_list(i_acquisition_time_line.Swst_changes_azimuthtimes),
        swst_changes_values_i=_pyxb_sequence_to_list(i_acquisition_time_line.Swst_changes_values),
        noise_packets_number_i=i_acquisition_time_line.noise_packets_number,
        noise_packets_azimuth_times_i=_pyxb_sequence_to_list(i_acquisition_time_line.noise_packets_azimuthtimes),
        internal_calibration_number_i=i_acquisition_time_line.Internal_calibration_number,
        internal_calibration_azimuth_times_i=_pyxb_sequence_to_list(
            i_acquisition_time_line.Internal_calibration_azimuthtimes),
        swl_changes_number_i=swl_changes_number,
        swl_changes_azimuth_times_i=_pyxb_sequence_to_list(i_acquisition_time_line.Swl_changes_azimuthtimes),
        swl_changes_values_i=_pyxb_sequence_to_list(i_acquisition_time_line.Swl_changes_values))

    if i_acquisition_time_line.DuplicatedLines_number is not None:
        o_acquisition_time_line.duplicated_lines = (
            i_acquisition_time_line.DuplicatedLines_number.value(),
            _pyxb_sequence_to_list(i_acquisition_time_line.DuplicatedLines_azimuthtimes))

    return o_acquisition_time_line


def convert_geo_point_to_aresys(i_point: pyxb_metadata_types.PointType) -> metadata.GeoPoint:
    return metadata.GeoPoint(*_pyxb_sequence_to_list(i_point))


def convert_ground_corner_points_to_aresys(
        i_ground_corner_points: pyxb_metadata_types.GroundCornersPointsType) -> metadata.GroundCornerPoints:
    o_ground_corner_points = metadata.GroundCornerPoints()
    o_ground_corner_points.easting_grid_size = i_ground_corner_points.EastingGridSize.value()
    o_ground_corner_points.northing_grid_size = i_ground_corner_points.NorthingGridSize.value()
    o_ground_corner_points.center_point = convert_geo_point_to_aresys(i_ground_corner_points.Center.Point)
    o_ground_corner_points.nw_point = convert_geo_point_to_aresys(i_ground_corner_points.NorthWest.Point)
    o_ground_corner_points.ne_point = convert_geo_point_to_aresys(i_ground_corner_points.NorthEast.Point)
    o_ground_corner_points.sw_point = convert_geo_point_to_aresys(i_ground_corner_points.SouthWest.Point)
    o_ground_corner_points.se_point = convert_geo_point_to_aresys(i_ground_corner_points.SouthEast.Point)

    return o_ground_corner_points


def convert_burst_to_aresys(i_burst: pyxb_metadata_types.BurstType, lines) -> metadata.Burst:
    o_burst = metadata.Burst(i_burst.RangeStartTime.value(),
                             PreciseDateTime().set_from_utc_string(i_burst.AzimuthStartTime.value()), lines,
                             i_burst.BurstCenterAzimuthShift)
    return o_burst


def convert_antenna_info_to_aresys(i_antenna_info: pyxb_metadata_types.AntennaInfoType) -> metadata.AntennaInfo:
    o_antenna_info = metadata.AntennaInfo(i_antenna_info.SensorName,
                                          i_antenna_info.Polarization,
                                          i_antenna_info.AcquisitionMode,
                                          i_antenna_info.BeamName)

    if i_antenna_info.LinesPerPattern is not None:
        o_antenna_info.lines_per_pattern = i_antenna_info.LinesPerPattern

    return o_antenna_info


def convert_data_statistics_to_aresys(
        i_data_statistics: pyxb_metadata_types.DataStatisticsType) -> metadata.DataStatistics:
    o_data_statistics = metadata.DataStatistics(i_data_statistics.NumSamples.value(),
                                                i_data_statistics.MaxI.value(),
                                                i_data_statistics.MaxQ.value(),
                                                i_data_statistics.MinI.value(),
                                                i_data_statistics.MinQ.value(),
                                                i_data_statistics.SumI.value(),
                                                i_data_statistics.SumQ.value(),
                                                i_data_statistics.Sum2I.value(),
                                                i_data_statistics.Sum2Q.value(),
                                                i_data_statistics.StdDevI.value(),
                                                i_data_statistics.StdDevQ.value())

    if i_data_statistics.StatisticsList is not None:
        for data_block_stat_pyxb in i_data_statistics.StatisticsList.DataBlockStatistic:
            data_block_stat = metadata.DataBlockStatistic(data_block_stat_pyxb.lineStart,
                                                          data_block_stat_pyxb.lineStop,
                                                          data_block_stat_pyxb.NumSamples.value(),
                                                          data_block_stat_pyxb.MaxI.value(),
                                                          data_block_stat_pyxb.MaxQ.value(),
                                                          data_block_stat_pyxb.MinI.value(),
                                                          data_block_stat_pyxb.MinQ.value(),
                                                          data_block_stat_pyxb.SumI.value(),
                                                          data_block_stat_pyxb.SumQ.value(),
                                                          data_block_stat_pyxb.Sum2I.value(),
                                                          data_block_stat_pyxb.Sum2Q.value())
            o_data_statistics.add_data_block_statistic(data_block_stat)

    return o_data_statistics


def convert_poly_2d_to_aresys(i_poly_2d, i_poly_type=None):

    if i_poly_type is None:
        poly_type = metadata._Poly2DVector
    else:
        poly_type = i_poly_type

    o_poly_2d_vector = poly_type()
    for current_poly in i_poly_2d:
        coefficients = np.zeros(len(current_poly.pol.val))
        for index_pol, value in enumerate(current_poly.pol.val):
            coefficients[index_pol] = value.value()
        o_poly_2d_vector.add_poly(
            poly_type.get_single_poly_type()(PreciseDateTime().set_from_utc_string(current_poly.taz0_Utc.value()),
                                             current_poly.trg0_s.value(),
                                             coefficients))

    return o_poly_2d_vector


def convert_doppler_centroid_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.DopplerCentroidVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.DopplerCentroidVector)


def convert_doppler_rate_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.DopplerRateVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.DopplerRateVector)


def convert_tops_azimuth_modulation_rate_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.TopsAzimuthModulationRateVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.TopsAzimuthModulationRateVector)


def convert_slant_to_ground_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.SlantToGroundVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.SlantToGroundVector)


def convert_ground_to_slant_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.GroundToSlantVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.GroundToSlantVector)


def convert_slant_to_incidence_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.SlantToIncidenceVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.SlantToIncidenceVector)


def convert_slant_to_elevation_to_aresys(
        i_poly: pyxb_metadata_types.polyType) -> metadata.SlantToElevationVector:
    return convert_poly_2d_to_aresys(i_poly, metadata.SlantToElevationVector)


_COREG_POLY_COEF_PYXB_TO_ARESYS = {0: 4,
                                   1: 1,
                                   2: 2,
                                   3: 0,
                                   }


def convert_coreg_poly_to_aresys(
        i_coreg_poly: pyxb_metadata_types.polyCoregType) -> metadata.CoregPolyVector:
    o_coreg_poly_vector = metadata.CoregPolyVector()
    for current_poly in i_coreg_poly:
        coefficients_rg = np.zeros(max(_COREG_POLY_COEF_PYXB_TO_ARESYS.values())+1)
        coefficients_az = np.zeros(max(_COREG_POLY_COEF_PYXB_TO_ARESYS.values())+1)
        for index_pyxb, index_aresys in _COREG_POLY_COEF_PYXB_TO_ARESYS.items():
            coefficients_rg[index_aresys] = \
                current_poly.polRg.val[index_pyxb].value()
            coefficients_az[index_aresys] = \
                current_poly.polAz.val[index_pyxb].value()
        o_coreg_poly_vector.add_poly(
            metadata.CoregPoly(PreciseDateTime().set_from_utc_string(current_poly.taz0_Utc.value()),
                               current_poly.trg0_s.value(),
                               coefficients_az,
                               coefficients_rg))

    return o_coreg_poly_vector


def _expand_lines_per_burst_dict(lines_per_burst, num_bursts):
    out = list()
    for i_burst in range(num_bursts):
        for i_burst_change in sorted(lines_per_burst.keys(), reverse=True):
            if i_burst + 1 >= i_burst_change:
                out.append(lines_per_burst[i_burst_change])
                break
    return out


def _reduce_to_lines_per_burst_dict(lines_per_burst):
    out = dict()
    last_num_lines = 0
    for i_burst, lines_per_burst in enumerate(lines_per_burst):
        if lines_per_burst != last_num_lines:
            last_num_lines = lines_per_burst
            out[i_burst] = last_num_lines
    return out


def convert_burst_info_to_aresys(i_burst_info: pyxb_metadata_types.BurstInfoType) -> metadata.BurstInfo:
    number_of_bursts = i_burst_info.NumberOfBursts

    o_burst_info = metadata.BurstInfo(burst_repetition_frequency=i_burst_info.BurstRepetitionFrequency.value())
    if i_burst_info.LinesPerBurst is not None:
        lines_per_burst = [i_burst_info.LinesPerBurst for _ in range(number_of_bursts)]
    else:
        lines_per_burst = dict()
        for lines in i_burst_info.LinesPerBurstChangeList.Lines:
            lines_per_burst[lines.FromBurst] = lines.value()
        lines_per_burst = _expand_lines_per_burst_dict(lines_per_burst, number_of_bursts)

    for i_burst in range(number_of_bursts):
        burst = convert_burst_to_aresys(i_burst_info.Burst[i_burst], lines_per_burst[i_burst])
        o_burst_info.add_burst(burst.range_start_time, burst.azimuth_start_time,
                               burst.lines, burst.burst_center_azimuth_shift)

    return o_burst_info


_SECTION_TO_CONVERTER_TO_ARESYS = {
    'SamplingConstants': convert_sampling_constants_to_aresys,
    'Pulse': convert_pulse_to_aresys,
    'SwathInfo': convert_swath_info_to_aresys,
    'RasterInfo': convert_rasterinfo_to_aresys,
    'DataSetInfo': convert_dataset_info_to_aresys,
    'StateVectorData': convert_state_vectors_to_aresys,
    'AttitudeInfo': convert_attitude_info_to_aresys,
    'AcquisitionTimeLine': convert_acquisition_time_line_to_aresys,
    'GroundCornerPoints': convert_ground_corner_points_to_aresys,
    'BurstInfo': convert_burst_info_to_aresys,
    'DopplerCentroid': convert_doppler_centroid_to_aresys,
    'DopplerRate': convert_doppler_rate_to_aresys,
    'TopsAzimuthModulationRate': convert_tops_azimuth_modulation_rate_to_aresys,
    'SlantToGround': convert_slant_to_ground_to_aresys,
    'GroundToSlant': convert_ground_to_slant_to_aresys,
    'SlantToIncidence': convert_slant_to_incidence_to_aresys,
    'SlantToElevation': convert_slant_to_elevation_to_aresys,
    'AntennaInfo': convert_antenna_info_to_aresys,
    'DataStatistics': convert_data_statistics_to_aresys,
    'CoregPoly': convert_coreg_poly_to_aresys,
}

_SECTION_ARESYS_TO_PYXB = {
    'SamplingConstants': 'SamplingConstants',
    'Pulse': 'Pulse',
    'SwathInfo': 'SwathInfo',
    'RasterInfo': 'RasterInfo',
    'DataSetInfo': 'DataSetInfo',
    'StateVectors': 'StateVectorData',
    'AttitudeInfo': 'AttitudeInfo',
    'AcquisitionTimeLine': 'AcquisitionTimeLine',
    'GroundCornerPoints': 'GroundCornerPoints',
    'BurstInfo': 'BurstInfo',
    'DopplerCentroidVector': 'DopplerCentroid',
    'DopplerRateVector': 'DopplerRate',
    'TopsAzimuthModulationRateVector': 'TopsAzimuthModulationRate',
    'SlantToGroundVector': 'SlantToGround',
    'GroundToSlantVector': 'GroundToSlant',
    'SlantToIncidenceVector': 'SlantToIncidence',
    'SlantToElevationVector': 'SlantToElevation',
    'AntennaInfo': 'AntennaInfo',
    'DataStatistics': 'DataStatistics',
    'CoregPolyVector': 'CoregPoly',
}

supported_elements = metadata.MetaDataChannel.get_supported_metadata_elements()
supported_elements.append('StateVectorData')  # in MetaDataChannel is StateVectors
readable_elements = list(_SECTION_TO_CONVERTER_TO_ARESYS.keys())
readable_elements.append('StateVectors')  # in MetaDataChannel is StateVectors
for supported_element in supported_elements:
    if supported_element not in readable_elements:
        logging.getLogger(__name__).warning("{} not readable".format(supported_element))
for readable_element in readable_elements:
    if readable_element not in supported_elements:
        logging.getLogger(__name__).warning("{} not supported".format(readable_element))


def run_conversion_to_aresys(metadata_i: pyxb_metadata.AresysXmlDocType):
    metadata_o = metadata.MetaData()

    for channel in metadata_i.Channel:
        mdc = metadata.MetaDataChannel()

        # RasterInfo
        if channel.RasterInfo is not None:
            raster_info = convert_rasterinfo_to_aresys(channel.RasterInfo)
            mdc.insert_element(raster_info)
        else:
            logger = logging.getLogger(__name__)
            logger.debug("Raster info not available in Channel #{}".format(channel.Number))

        for section, converter in _SECTION_TO_CONVERTER_TO_ARESYS.items():
            if section != "RasterInfo":
                section_pyxb = getattr(channel, section)
                if section_pyxb is not None:
                    mdc.insert_element(converter(section_pyxb))

        metadata_o.append_channel(mdc)

    return metadata_o


def convert_sampling_constants_to_pyxb(
        i_sampling_constants: metadata.SamplingConstants) -> pyxb_metadata_types.SamplingConstantsType:
    o_sampling_constants = pyxb_metadata_types.SamplingConstantsType(frg_hz=i_sampling_constants.frg_hz,
                                                                     Brg_hz=i_sampling_constants.brg_hz,
                                                                     faz_hz=i_sampling_constants.faz_hz,
                                                                     Baz_hz=i_sampling_constants.baz_hz)

    o_sampling_constants.frg_hz.unit = str(i_sampling_constants.frg_hz_unit)
    o_sampling_constants.Brg_hz.unit = str(i_sampling_constants.brg_hz_unit)
    o_sampling_constants.faz_hz.unit = str(i_sampling_constants.faz_hz_unit)
    o_sampling_constants.Baz_hz.unit = str(i_sampling_constants.baz_hz_unit)

    return o_sampling_constants


def convert_pulse_to_pyxb(i_pulse: metadata.Pulse):
    o_pulse = pyxb_metadata_types.PulseType(PulseLength=i_pulse.pulse_length,
                                            Bandwidth=i_pulse.bandwidth,
                                            PulseSamplingRate=i_pulse.pulse_sampling_rate,
                                            PulseEnergy=i_pulse.pulse_energy)

    if i_pulse.pulse_direction is not None:
        o_pulse.Direction = i_pulse.pulse_direction.value
    if i_pulse.pulse_start_frequency is not None:
        o_pulse.PulseStartFrequency = i_pulse.pulse_start_frequency
        o_pulse.PulseStartFrequency.unit = str(i_pulse.pulse_start_frequency_unit)
    if i_pulse.pulse_start_phase is not None:
        o_pulse.PulseStartPhase = i_pulse.pulse_start_phase
        o_pulse.PulseStartPhase.unit = str(i_pulse.pulse_start_phase_unit)

    o_pulse.PulseLength.unit = str(i_pulse.pulse_length_unit)
    o_pulse.Bandwidth.unit = str(i_pulse.bandwidth_unit)
    o_pulse.PulseEnergy.unit = str(i_pulse.pulse_energy_unit)
    o_pulse.PulseSamplingRate.unit = str(i_pulse.pulse_sampling_rate_unit)

    return o_pulse


def convert_swath_info_to_pyxb(i_swath_info: metadata.SwathInfo) -> pyxb_metadata_types.SwathInfoType:
    o_swath_info = pyxb_metadata_types.SwathInfoType(
        Swath=i_swath_info.swath,
        SwathAcquisitionOrder=i_swath_info.swath_acquisition_order,
        Polarization=i_swath_info.polarization.value,
        Rank=i_swath_info.rank,
        RangeDelayBias=i_swath_info.range_delay_bias,
        AcquisitionStartTime=str(i_swath_info.acquisition_start_time),
        AzimuthSteeringRateReferenceTime=i_swath_info.azimuth_steering_rate_reference_time,
        AzimuthSteeringRatePol=pyxb.BIND(),
        AcquisitionPRF=i_swath_info.acquisition_prf,
        EchoesPerBurst=i_swath_info.echoes_per_burst,
    )

    for index, val in enumerate(i_swath_info.azimuth_steering_rate_pol):
        o_swath_info.AzimuthSteeringRatePol.append(val)
        o_swath_info.AzimuthSteeringRatePol.val[index].N = index+1

    if i_swath_info.channel_delay is not None:
        o_swath_info.ChannelDelay = i_swath_info.channel_delay
    if i_swath_info.rx_gain is not None:
        o_swath_info.RxGain = i_swath_info.rx_gain

    o_swath_info.AcquisitionPRF.unit = str(i_swath_info.acquisition_prf_unit)
    o_swath_info.RangeDelayBias.unit = str(i_swath_info.range_delay_bias_unit)
    o_swath_info.AcquisitionStartTime.unit = str(i_swath_info.acquisition_start_time_unit)
    o_swath_info.AzimuthSteeringRateReferenceTime.unit = str(i_swath_info.az_steering_rate_ref_time_unit)

    return o_swath_info


def convert_rasterinfo_to_pyxb(i_rasterinfo: metadata.RasterInfo):
    o_rasterinfo = pyxb_metadata_types.RasterInfoType(Lines=i_rasterinfo.lines,
                                                      LinesStep=i_rasterinfo.lines_step,
                                                      LinesStart=str(i_rasterinfo.lines_start),
                                                      Samples=i_rasterinfo.samples,
                                                      SamplesStep=i_rasterinfo.samples_step,
                                                      SamplesStart=str(i_rasterinfo.samples_start),
                                                      HeaderOffsetBytes=i_rasterinfo.header_offset_bytes,
                                                      RowPrefixBytes=i_rasterinfo.row_prefix_bytes,
                                                      FileName=i_rasterinfo.file_name,
                                                      ByteOrder=i_rasterinfo.byte_order.value,
                                                      CellType=i_rasterinfo.cell_type.value)

    o_rasterinfo.SamplesStart.unit = str(i_rasterinfo.samples_start_unit)
    o_rasterinfo.SamplesStep.unit = str(i_rasterinfo.samples_step_unit)
    o_rasterinfo.LinesStart.unit = str(i_rasterinfo.lines_start_unit)
    o_rasterinfo.LinesStep.unit = str(i_rasterinfo.lines_step_unit)
    if i_rasterinfo.invalid_value is not None:
        o_rasterinfo.InvalidValue = pyxb_metadata_types.DCOMPLEX(RealValue=i_rasterinfo.invalid_value.real,
                                                                 ImaginaryValue=i_rasterinfo.invalid_value.imag)
    if i_rasterinfo.format_type is not None:
        o_rasterinfo.RasterFormat = i_rasterinfo.format_type.value

    return o_rasterinfo


def convert_dataset_info_to_pyxb(i_dataset_info: metadata.DataSetInfo) -> pyxb_metadata_types.DataSetInfoType:
    if i_dataset_info.sense_date is not None:
        sense_date_str = str(i_dataset_info.sense_date)
    else:
        sense_date_str = 'NOT_AVAILABLE'

    if i_dataset_info.processing_date is not None:
        processing_date_str = str(i_dataset_info.processing_date)
    else:
        processing_date_str = 'NOT_AVAILABLE'

    o_dataset_info = pyxb_metadata_types.DataSetInfoType(
        SensorName=i_dataset_info.sensor_name,
        Description=i_dataset_info.description,
        SenseDate=sense_date_str,
        AcquisitionMode=i_dataset_info.acquisition_mode,
        ImageType=i_dataset_info.image_type,
        Projection=i_dataset_info.projection,
        AcquisitionStation=i_dataset_info.acquisition_station,
        ProcessingCenter=i_dataset_info.processing_center,
        ProcessingDate=processing_date_str,
        ProcessingSoftware=i_dataset_info.processing_software,
        fc_hz=i_dataset_info.fc_hz,
        SideLooking=i_dataset_info.side_looking.value,
        ExternalCalibrationFactor=i_dataset_info.external_calibration_factor,
        DataTakeID=i_dataset_info.data_take_id)
    return o_dataset_info


def convert_state_vectors_to_pyxb(i_state_vectors: metadata.StateVectors) -> pyxb_metadata_types.StateVectorDataType:
    o_state_vectors = pyxb_metadata_types.StateVectorDataType(
        pSV_m=pyxb.BIND(),
        vSV_mOs=pyxb.BIND(),
        OrbitNumber='NOT_AVAILABLE' if i_state_vectors.orbit_number < 0 else str(i_state_vectors.orbit_number),
        Track='NOT_AVAILABLE' if i_state_vectors.track_number < 0 else str(i_state_vectors.track_number),
        OrbitDirection=i_state_vectors.orbit_direction.value,
        t_ref_Utc=str(i_state_vectors.reference_time),
        dtSV_s=str(i_state_vectors.time_step),
        nSV_n=i_state_vectors.number_of_state_vectors)

    for index_position, position in enumerate(i_state_vectors.position_vector):
        for index_sv_vect, coord in enumerate(position):
            index_current = index_position * 3 + index_sv_vect + 1

            o_state_vectors.pSV_m.append(coord)
            o_state_vectors.pSV_m.val[index_current - 1].N = str(index_current)

            o_state_vectors.vSV_mOs.append(i_state_vectors.velocity_vector[index_position][index_sv_vect])
            o_state_vectors.vSV_mOs.val[index_current - 1].N = str(index_current)

    return o_state_vectors


def _list_to_pyxb_sequence(input_list, pyxb_obj, attributes=None):
    if input_list is None:
        return
    if attributes is None:
        attributes = {}
    if len(input_list) > 0:
        for i_val, val in enumerate(input_list):
            pyxb_obj.append(val)
            pyxb_obj.val[i_val].N = i_val + 1

            for attr_name, attr_value in attributes.items():
                setattr(pyxb_obj.val[i_val], attr_name, attr_value)


def convert_attitude_info_to_pyxb(i_attitude_info: metadata.AttitudeInfo) -> pyxb_metadata_types.AttitudeInfoType:
    o_attitude_info = pyxb_metadata_types.AttitudeInfoType(
        yaw_deg=pyxb.BIND(),
        pitch_deg=pyxb.BIND(),
        roll_deg=pyxb.BIND(),
        t_ref_Utc=str(i_attitude_info.reference_time),
        dtYPR_s=str(i_attitude_info.time_step),
        nYPR_n=i_attitude_info.attitude_records_number,
        referenceFrame=i_attitude_info.reference_frame.value.upper(),
        rotationOrder=i_attitude_info.rotation_order.value.upper(),
        AttitudeType=i_attitude_info.attitude_type.value.upper())

    _list_to_pyxb_sequence(i_attitude_info.yaw_vector, o_attitude_info.yaw_deg)
    _list_to_pyxb_sequence(i_attitude_info.pitch_vector, o_attitude_info.pitch_deg)
    _list_to_pyxb_sequence(i_attitude_info.roll_vector, o_attitude_info.roll_deg)

    return o_attitude_info


def convert_acquisition_time_line_to_pyxb(
        i_acquisition_time_line: metadata.AcquisitionTimeLine) -> pyxb_metadata_types.AcquisitionTimelineType:
    o_acquisition_time_line = pyxb_metadata_types.AcquisitionTimelineType(
        MissingLines_number=i_acquisition_time_line.missing_lines[0],
        MissingLines_azimuthtimes=pyxb.BIND(),
        Swst_changes_number=i_acquisition_time_line.swst_changes[0],
        Swst_changes_azimuthtimes=pyxb.BIND(),
        Swst_changes_values=pyxb.BIND(),
        noise_packets_number=i_acquisition_time_line.noise_packet[0],
        noise_packets_azimuthtimes=pyxb.BIND(),
        Internal_calibration_number=i_acquisition_time_line.internal_calibration[0],
        Internal_calibration_azimuthtimes=pyxb.BIND(),
        Swl_changes_number=i_acquisition_time_line.swl_changes[0],
        Swl_changes_azimuthtimes=pyxb.BIND(),
        Swl_changes_values=pyxb.BIND(),
        DuplicatedLines_number=i_acquisition_time_line.duplicated_lines[0],
        DuplicatedLines_azimuthtimes=pyxb.BIND())

    _list_to_pyxb_sequence(i_acquisition_time_line.missing_lines[1], o_acquisition_time_line.MissingLines_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.missing_lines_azimuth_times_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.swst_changes[1], o_acquisition_time_line.Swst_changes_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.swst_changes_azimuth_times_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.swst_changes[2], o_acquisition_time_line.Swst_changes_values,
                           attributes={'unit': i_acquisition_time_line.swst_changes_values_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.noise_packet[1], o_acquisition_time_line.noise_packets_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.noise_packets_azimuth_times_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.internal_calibration[1],
                           o_acquisition_time_line.Internal_calibration_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.internal_calibration_azimuth_times_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.swl_changes[1], o_acquisition_time_line.Swl_changes_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.swl_changes_azimuth_times_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.swl_changes[2], o_acquisition_time_line.Swl_changes_values,
                           attributes={'unit': i_acquisition_time_line.swl_changes_values_unit})
    _list_to_pyxb_sequence(i_acquisition_time_line.duplicated_lines[1],
                           o_acquisition_time_line.DuplicatedLines_azimuthtimes,
                           attributes={'unit': i_acquisition_time_line.duplicated_lines_azimuth_times_unit})

    return o_acquisition_time_line


def convert_geo_point_to_pyxb(i_geo_point: metadata.GeoPoint) -> pyxb_metadata_types.PointType:
    o_point = pyxb_metadata_types.PointType()

    for i_val, val in enumerate(i_geo_point.to_list()):
        o_point.append(val)
        o_point.val[i_val].N = i_val + 1

    return o_point


def convert_ground_corner_points_to_pyxb(
        i_ground_corner_points: metadata.GroundCornerPoints) -> pyxb_metadata_types.GroundCornersPointsType:
    o_ground_corner_points = pyxb_metadata_types.GroundCornersPointsType(
        EastingGridSize=i_ground_corner_points.easting_grid_size,
        NorthingGridSize=i_ground_corner_points.northing_grid_size,
        NorthWest=pyxb.BIND(convert_geo_point_to_pyxb(i_ground_corner_points.nw_point)),
        NorthEast=pyxb.BIND(convert_geo_point_to_pyxb(i_ground_corner_points.ne_point)),
        SouthWest=pyxb.BIND(convert_geo_point_to_pyxb(i_ground_corner_points.sw_point)),
        SouthEast=pyxb.BIND(convert_geo_point_to_pyxb(i_ground_corner_points.se_point)),
        Center=pyxb.BIND(convert_geo_point_to_pyxb(i_ground_corner_points.center_point))
    )
    return o_ground_corner_points


def convert_burst_to_pyxb(i_burst: metadata.Burst) -> pyxb_metadata_types.BurstType:
    o_burst = pyxb_metadata_types.BurstType(RangeStartTime=i_burst.range_start_time,
                                            AzimuthStartTime=str(i_burst.azimuth_start_time),
                                            BurstCenterAzimuthShift=i_burst.burst_center_azimuth_shift)
    o_burst.RangeStartTime.unit = constants.SECOND_STR
    o_burst.AzimuthStartTime.unit = constants.UTC_STR
    if o_burst.BurstCenterAzimuthShift is not None:
        o_burst.BurstCenterAzimuthShift.unit = constants.SECOND_STR
    return o_burst


def convert_burst_info_to_pyxb(i_burst_info: metadata.BurstInfo) -> pyxb_metadata_types.BurstInfoType:
    o_burst_info = pyxb_metadata_types.BurstInfoType(NumberOfBursts=i_burst_info.get_number_of_bursts(),
                                                     BurstRepetitionFrequency=i_burst_info.burst_repetition_frequency)

    o_burst_info.BurstRepetitionFrequency.unit = constants.HERTZ_STR

    lines_per_burst = list()
    for i_burst in range(i_burst_info.get_number_of_bursts()):
        lines_per_burst.append(i_burst_info.get_lines(i_burst))
    from_burst_dict = _reduce_to_lines_per_burst_dict(lines_per_burst)

    o_burst_info.LinesPerBurstChangeList = pyxb.BIND()

    for i_from_burst, (from_burst, lines) in enumerate(from_burst_dict.items()):
        o_burst_info.LinesPerBurstChangeList.Lines.append(lines)
        o_burst_info.LinesPerBurstChangeList.Lines[i_from_burst].FromBurst = from_burst + 1

    for i_burst in range(i_burst_info.get_number_of_bursts()):
        burst = i_burst_info.get_burst(i_burst)
        o_burst_info.Burst.append(convert_burst_to_pyxb(burst))
        o_burst_info.Burst[i_burst].N = i_burst + 1

    return o_burst_info


def convert_antenna_info_to_pyxb(i_antenna_info: metadata.AntennaInfo) -> pyxb_metadata_types.AntennaInfoType:
    o_antenna_info = pyxb_metadata_types.AntennaInfoType(SensorName=i_antenna_info.sensor_name,
                                                         AcquisitionMode=i_antenna_info.acquisition_mode,
                                                         BeamName=i_antenna_info.acquisition_beam,
                                                         Polarization=i_antenna_info.polarization.value,
                                                         )
    if i_antenna_info.lines_per_pattern != 0:
        o_antenna_info.LinesPerPattern = i_antenna_info.lines_per_pattern

    return o_antenna_info


def convert_data_block_statistics_to_pyxb(
        i_data_block_statistics: metadata.DataBlockStatistic) -> pyxb_metadata_types.DataBlockStatisticsType:
    o_data_block_statistics = pyxb_metadata_types.DataBlockStatisticsType(
        NumSamples=i_data_block_statistics.num_samples,
        MaxI=i_data_block_statistics.max_i,
        MinI=i_data_block_statistics.min_i,
        MaxQ=i_data_block_statistics.max_q,
        MinQ=i_data_block_statistics.min_q,
        SumI=i_data_block_statistics.sum_i,
        SumQ=i_data_block_statistics.sum_q,
        Sum2I=i_data_block_statistics.sum_2_i,
        Sum2Q=i_data_block_statistics.sum_2_q,
    )
    o_data_block_statistics.lineStart = i_data_block_statistics.line_start
    o_data_block_statistics.lineStop = i_data_block_statistics.line_stop

    return o_data_block_statistics


def convert_data_statistics_to_pyxb(
        i_data_statistics: metadata.DataStatistics) -> pyxb_metadata_types.DataStatisticsType:
    o_data_block_statistics = pyxb_metadata_types.DataStatisticsType(NumSamples=i_data_statistics.num_samples,
                                                                     MaxI=i_data_statistics.max_i,
                                                                     MinI=i_data_statistics.min_i,
                                                                     MaxQ=i_data_statistics.max_q,
                                                                     MinQ=i_data_statistics.min_q,
                                                                     SumI=i_data_statistics.sum_i,
                                                                     SumQ=i_data_statistics.sum_q,
                                                                     Sum2I=i_data_statistics.sum_2_i,
                                                                     Sum2Q=i_data_statistics.sum_2_q,
                                                                     StdDevI=i_data_statistics.std_dev_i,
                                                                     StdDevQ=i_data_statistics.std_dev_q,
                                                                     )
    if i_data_statistics.get_number_of_data_block_statistic() != 0:
        for index in range(0, i_data_statistics.get_number_of_data_block_statistic()-1):
            o_data_block_statistics.StatisticsList.DataBlockStatistic[index] = \
                convert_data_block_statistics_to_pyxb(i_data_statistics.get_data_block_statistic(index))

    return o_data_block_statistics


def convert_poly_2d_to_pyxb(i_poly_2d_vector: metadata._Poly2DVector) -> pyxb_metadata_types.polyType:
    o_poly_2d = list()
    units = metadata._Poly2D.get_units()

    for index_poly, poly_2d in enumerate(i_poly_2d_vector):
        current_poly = pyxb_metadata_types.polyType(trg0_s=poly_2d.t_ref_rg,
                                                    taz0_Utc=str(poly_2d.t_ref_az),
                                                    pol=pyxb.BIND())
        for index_coef, coef in enumerate(poly_2d.coefficients):
            current_poly.pol.val.append(coef)
            current_poly.pol.val[index_coef].N = index_coef + 1
            current_poly.pol.val[index_coef].unit = units[index_coef]

        current_poly.Number = index_poly + 1
        current_poly.Total = i_poly_2d_vector.get_number_of_poly()

        o_poly_2d.append(current_poly)

    return o_poly_2d


def convert_coreg_poly_to_pyxb(
        i_coreg_poly_vector: metadata.CoregPolyVector) -> pyxb_metadata_types.polyCoregType:
    o_coreg_poly = list()

    for index_poly, coreg_poly in enumerate(i_coreg_poly_vector):
        current_poly = pyxb_metadata_types.polyCoregType(trg0_s=coreg_poly.ref_range_time,
                                                         taz0_Utc=str(coreg_poly.ref_azimuth_time),
                                                         polRg=pyxb.BIND(),
                                                         polAz=pyxb.BIND())
        index_coef = 0
        for _, index_aresys in _COREG_POLY_COEF_PYXB_TO_ARESYS.items():
            current_poly.polAz.val.append(coreg_poly.azimuth_poly.coefficients[index_aresys])
            current_poly.polRg.val.append(coreg_poly.range_poly.coefficients[index_aresys])
            current_poly.polAz.val[index_coef].N = index_coef + 1
            current_poly.polRg.val[index_coef].N = index_coef + 1
            index_coef += 1

        current_poly.Number = index_poly + 1
        current_poly.Total = i_coreg_poly_vector.get_number_of_poly()

        o_coreg_poly.append(current_poly)

    return o_coreg_poly


_SECTION_TO_CONVERTER_TO_PYXB = {
    'SamplingConstants': convert_sampling_constants_to_pyxb,
    'Pulse': convert_pulse_to_pyxb,
    'SwathInfo': convert_swath_info_to_pyxb,
    'RasterInfo': convert_rasterinfo_to_pyxb,
    'DataSetInfo': convert_dataset_info_to_pyxb,
    'StateVectors': convert_state_vectors_to_pyxb,
    'AttitudeInfo': convert_attitude_info_to_pyxb,
    'AcquisitionTimeLine': convert_acquisition_time_line_to_pyxb,
    'GroundCornerPoints': convert_ground_corner_points_to_pyxb,
    'BurstInfo': convert_burst_info_to_pyxb,
    'DopplerCentroidVector': convert_poly_2d_to_pyxb,
    'DopplerRateVector': convert_poly_2d_to_pyxb,
    'TopsAzimuthModulationRateVector': convert_poly_2d_to_pyxb,
    'SlantToGroundVector': convert_poly_2d_to_pyxb,
    'GroundToSlantVector': convert_poly_2d_to_pyxb,
    'SlantToIncidenceVector': convert_poly_2d_to_pyxb,
    'SlantToElevationVector': convert_poly_2d_to_pyxb,
    'AntennaInfo': convert_antenna_info_to_pyxb,
    'DataStatistics': convert_data_statistics_to_pyxb,
    'CoregPolyVector': convert_coreg_poly_to_pyxb,
}

supported_elements = metadata.MetaDataChannel.get_supported_metadata_elements()
writable_elements = list(_SECTION_TO_CONVERTER_TO_PYXB.keys())
for supported_element in supported_elements:
    if supported_element not in writable_elements:
        logging.getLogger(__name__).warning("{} not writable".format(supported_element))
for writable_element in writable_elements:
    if writable_element not in supported_elements:
        logging.getLogger(__name__).warning("{} not supported".format(writable_element))


def run_conversion_pyxb(metadata_i: metadata.MetaData) -> pyxb_metadata.AresysXmlDocType:
    version = 2.1
    o_metadata = pyxb_metadata.AresysXmlDocType(metadata_i.get_number_of_channels(), version, '')

    for index_channel in range(0, metadata_i.get_number_of_channels()):
        o_metadata.Channel.append(pyxb.BIND())
        o_metadata.Channel[index_channel].Number = index_channel + 1
        o_metadata.Channel[index_channel].Total = metadata_i.get_number_of_channels()

        # RasterInfo is mandatory
        raster_info_aresys = metadata_i.get_metadata_channels(index_channel).get_element('RasterInfo')
        o_metadata.Channel[index_channel].RasterInfo = convert_rasterinfo_to_pyxb(raster_info_aresys)

        # Other sections
        for section, converter in _SECTION_TO_CONVERTER_TO_PYXB.items():
            if section != "RasterInfo":
                section_aresys = metadata_i.get_metadata_channels(index_channel).get_element(section)
                if section_aresys is not None:
                    setattr(o_metadata.Channel[index_channel],
                            _SECTION_ARESYS_TO_PYXB[section], converter(section_aresys))

    return o_metadata



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import logging
import numpy as np
from scipy.interpolate import griddata
from biopal.utility.utility_functions import epsg_in_to_epsg_out
from biopal.utility.constants import (
    EPSG_CODE_ECEF,
    EPSG_CODE_LLA,
    MAX_EARTH_RADIUS_METERS,
)


def geocoding_init(ecef_grid, rg_vec_subs, az_vec_subs, min_spacing_m):

    ### Convert ECEFGRID pixel to pixel from X+Y+Z to L+L+H (with for example epsg_in_to_epsg_out): the result will NOT be a uniform grid of lat lon
    logging.info("convert ECEFGRID pixel to pixel from X Y Z to Lat Lon Alt...")
    lon_in, lat_in, alt_in = epsg_in_to_epsg_out(
        ecef_grid["X"][rg_vec_subs, :][:, az_vec_subs],
        ecef_grid["Y"][rg_vec_subs, :][:, az_vec_subs],
        ecef_grid["Z"][rg_vec_subs, :][:, az_vec_subs],
        EPSG_CODE_ECEF,
        EPSG_CODE_LLA,
    )

    lat_lon_step = np.rad2deg(min_spacing_m / MAX_EARTH_RADIUS_METERS)

    logging.info("    geocoding latitude and longitude step used: {} [deg]".format(lat_lon_step))
    lat_lon_gap = 0  # 1*lat_lon_step

    # latitude should be decrescent
    used_lat_step = -np.abs(lat_lon_step)
    lat_regular_vector = np.arange(np.nanmax(lat_in) - lat_lon_gap, np.nanmin(lat_in) + lat_lon_gap, used_lat_step)
    # longitude should be crescent
    used_lon_step = np.abs(lat_lon_step)
    lon_regular_vector = np.arange(np.nanmin(lon_in) - lat_lon_gap, np.nanmax(lon_in) + lat_lon_gap, used_lon_step)

    lonMeshed_out, latMeshed_out = np.meshgrid(lon_regular_vector, lat_regular_vector)

    valid_values_mask = np.invert(np.isnan(lat_in) + np.isnan(lon_in) + np.isnan(alt_in))

    return (
        lon_in,
        lat_in,
        alt_in,
        lonMeshed_out,
        latMeshed_out,
        valid_values_mask,
        lon_regular_vector,
        used_lon_step,
        lat_regular_vector,
        used_lat_step,
    )


def geocoding(
    data_to_geocode, lon_in, lat_in, lonMeshed_out, latMeshed_out, valid_values_mask, interp_method="linear",
):
    f = griddata(
        np.array([lat_in[valid_values_mask].flatten(), lon_in[valid_values_mask].flatten()]).T,
        data_to_geocode[valid_values_mask].flatten(),
        np.array([latMeshed_out.flatten(), lonMeshed_out.flatten()]).T,
        method=interp_method,
    )  # interpolate to fill holes
    data_ground = f.reshape(latMeshed_out.shape)

    return data_ground



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Generic polynomial module
-------------------------
"""

from arepytools.io.metadata import _Poly2D
from arepytools.io.metadata import _Poly2DVector


class GenericPoly:
    """
    Generic polynomial class
    """

    def __init__(self, reference_values, coefficients, powers):
        """
        Initialize the GenericPoly class

        :param reference_values:
        :param coefficients:
        :param powers:
        """

        self.reference_values = reference_values
        self.poly = dict()
        for index_coefficients, coefficient in enumerate(coefficients):
            self.poly[coefficient] = powers[index_coefficients]

    def __repr__(self):
        representation = "coefficients: {}\n".format(list(self.poly.keys()))
        representation += "powers: {}\n".format(list(self.poly.values()))
        representation += "reference values: {}\n".format(list(self.reference_values))
        return representation

    def evaluate(self, values):
        """
        Evaluate the GenericPoly for the values provided

        :param values:
        :return:
        """
        values = tuple(values)
        result = 0
        for coefficient, powers in self.poly.items():
            current_result = 1
            for index_dimensions, ref_val in enumerate(self.reference_values):
                current_result *= \
                    (values[index_dimensions] - ref_val) ** powers[index_dimensions]
            result += coefficient * current_result

        return result


class SortedPolyList:
    """
    SortedPolyList class
    """

    def __init__(self, reference_index=0, list_generic_poly=None):
        """
        Initialize sorted poly list

        :param reference_index:
        :param list_generic_poly:
        """
        self.reference_index = reference_index
        if list_generic_poly is not None:
            self._sorted_poly_list = list_generic_poly
            self._sort_poly_list()
        else:
            self._sorted_poly_list = list()

    def __repr__(self):
        representation = "Sorted Poly List:\n"
        for index, poly in enumerate(self._sorted_poly_list):
            representation += "Poly #{}: \n{}\n".format(index, poly)
        return representation

    def append(self, generic_poly):
        """
        Append a new poly to the SortedPolyList

        :param generic_poly:
        """
        self._sorted_poly_list.append(generic_poly)
        self._sort_poly_list()

    def _sort_poly_list(self):
        self._sorted_poly_list.sort(key=lambda x: x.reference_values[self.reference_index])

    def evaluate(self, values):
        """
        Evaluate the SortedPolyList for the values provided

        :param values:
        :return:
        """
        previous_poly = self._sorted_poly_list[0]
        for poly in self._sorted_poly_list:
            if poly.reference_values[self.reference_index] > values[self.reference_index]:
                break
            previous_poly = poly
        return previous_poly.evaluate(values)


def _create_generic_poly(poly2d:_Poly2D):
    """
    Create a GenericPoly from the values provided in metadata classes of base type _Poly2D

    :param poly2d:
    :return generic_poly:
    """
    return GenericPoly(
        reference_values=[poly2d.t_ref_az, poly2d.t_ref_rg],
        coefficients=poly2d.coefficients,
        powers=list(zip(poly2d.get_powers_x(), poly2d.get_powers_y()))
    )

def create_sorted_poly_list(poly2d_vector:_Poly2DVector):
    """
    Create a SortedPolyList from the values provided in metadata classes of base type _Poly2DVector

    :param poly2d_vector:
    :return sorted_poly_list:
    """
    return SortedPolyList(list_generic_poly=[_create_generic_poly(p) for p in poly2d_vector])



from biopal.ground_cancellation.ground_cancellation import ground_cancellation
import numpy as np
import warnings


class TestGroundCancellation():
    """
    "ground_cancellation" APP functional tests
    performed with pytest framework: https://docs.pytest.org/en/stable/getting-started.html
    
    "ground_cancellation" APP executed twice:
        
        first with space varying feature disabled and dummy inputs.
        
        second with space varying feature enabled and dummy inputs: 
               to enable the space varying feature, the scalar enhanced forest height 
               coming from the "initial_values" function is converted to a 2D map.
        
    It checks the output to be a dictionary with the three polarizations, each 
    containing a 2D map with expected Nrg, Naz dimensions  in addition with 
    expected data type.
   
    Test execution:
    "pytest" needs to be installed in the environment:
        
        >>>pip install -U pytest
        
    From command line go in the "BioPAL/" directory and type the following instruction:
        
        >>>pytest
        
    This will automatically execute all the pytest-formatted tests in BioPAL
    
    Parameters
    ----------
    None, all the dummy parameters are loaded from the "initial_values" function
    
    """

    def test_ground_cancellation(self):
        """
        ground_cancellation functional test #1
        Executes ground_cancellation APP, with space varying feature disabled and 
        dummy inputs
        """
        
        (
         Nrg, 
         Naz,            
         data_stack_dict,
         wave_number_dict,
         multi_master_flag,
         enhanced_forest_height,
         equalization_flag,
         resolution_m_slant_rg,
         off_nadir_angle,
         slope,
         ) = self.initial_values()
        
        notched_pol_dict = ground_cancellation(
            data_stack_dict,
            wave_number_dict,
            multi_master_flag,
            enhanced_forest_height,
            equalization_flag,
            resolution_m_slant_rg,
            off_nadir_angle,
            slope,
        )
        
        pols = notched_pol_dict.keys()
        assert ('hh' in pols and 'hv' in pols and 'vv' in pols) and len(pols)==3, 'wrong output number of polarizations, in test with space varying OFF'
            
        for pol, data in notched_pol_dict.items():
            
            assert data.shape==(Nrg,Naz), 'wrong output shape, in test with space varying OFF'
            
            if multi_master_flag:
                assert isinstance(data[0,0], np.float64), 'wrong output data type; when multi_master_flag is true it is supposed to be float64, in test with space varying OFF'
            else:
                assert isinstance(data[0,0], np.complex128), 'wrong output data type; when multi_master_flag is false it is supposed to be complex128, in test with space varying OFF'
    
    
    def test_ground_cancellation_space_varying(self):
        """
        ground_cancellation functional test #2
        Executes ground_cancellation APP, with space varying feature enabled and 
        dummy inputs
        """
        
        (
          Nrg, 
          Naz,            
          data_stack_dict,
          wave_number_dict,
          multi_master_flag,
          enhanced_forest_height_scalar,
          equalization_flag,
          resolution_m_slant_rg,
          off_nadir_angle,
          slope,
          ) = self.initial_values()
        
        # converting to 2D map, in order to enable the space varying feature
        enhanced_forest_height_map = slope.copy()
        enhanced_forest_height_map.fill(enhanced_forest_height_scalar)
        
        warnings.filterwarnings('ignore') # warnings silenced for convenience
        notched_pol_dict = ground_cancellation(
            data_stack_dict,
            wave_number_dict,
            multi_master_flag,
            enhanced_forest_height_map,
            equalization_flag,
            resolution_m_slant_rg,
            off_nadir_angle,
            slope,
        )
        
        pols = notched_pol_dict.keys()
        assert ('hh' in pols and 'hv' in pols and 'vv' in pols) and len(pols)==3, 'wrong output number of polarizations, in test with space varying ON'
            
        for pol, data in notched_pol_dict.items():
            
            assert data.shape==(Nrg,Naz), 'wrong output shape, in test with space varying ON'
            
            if multi_master_flag:
                assert isinstance(data[0,0], np.float64), 'wrong output data type; when multi_master_flag is true it is supposed to be float64, in test with space varying ON'
            else:
                assert isinstance(data[0,0], np.complex128), 'wrong output data type; when multi_master_flag is false it is supposed to be complex128, in test with space varying ON'
    
    
    def initial_values(self):
        """
        Generates dummy variables for the ground cancellation APP, for functional testing
        
        Returns
        -------
        Nrg
            integer, number of range samples for the dummy maps
        Naz
            integer, number of azimuth lines for the dummy maps
        data_stack_dict
            dictionary with three acquisitions, each containing a sub dictionary with 
            three polarizations, each containing  a raster with random complex values
        wave_number_dict
            dictionary with three acquisitions, each containing a raster with 
            constant dummy real values
        multi_master_flag
            boolean (see ground cancellation documentation)
        enhanced_forest_height
                positive number (see ground cancellation documentation)
        equalization_flag
            string, enumeration flag (see ground cancellation documentation)
        resolution_m_slant_rg
           positive number, slant range resolutuon 
        off_nadir_angle
            map with dummy constant values for off nadir angle (supposed radiants)
        slope
            map with dummy constant values for slope angle (supposed radiants)
        """
        
        Nrg = 378
        Naz = 2656
        data_dummy = 1/np.sqrt(2)*np.random.rand(Nrg,Naz)+ 1j/np.sqrt(2)*np.random.rand(Nrg,Naz)
        pol_dict = {'hh': data_dummy, 'hv': data_dummy, 'vv': data_dummy }
        data_stack_dict = {'BSL_00': pol_dict, 'BSL_01': pol_dict, 'BSL_02': pol_dict }
        
        wave_number_dummy = np.ones((Nrg,Naz))
        wave_number_dict = {'BSL_00': 0*wave_number_dummy, 'BSL_01': -0.03*wave_number_dummy, 'BSL_02': 0.08*wave_number_dummy }
        
        multi_master_flag = True
        enhanced_forest_height = 50
        equalization_flag = '1'
        resolution_m_slant_rg = 25
        off_nadir_angle = 0.5*np.ones((Nrg,Naz))
        slope = 0.05*np.ones((Nrg,Naz))
        
        return (
            Nrg, 
            Naz,            
            data_stack_dict,
            wave_number_dict,
            multi_master_flag,
            enhanced_forest_height,
            equalization_flag,
            resolution_m_slant_rg,
            off_nadir_angle,
            slope,
            )        
        


# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Axis module
-----------
"""

import os
import typing
import numpy as np
import scipy.interpolate

from arepytools import _utils
from arepytools.timing.precisedatetime import PreciseDateTime

# Types
RealNumber = typing.Union[int, float]
AxisStartType = typing.Union[RealNumber, PreciseDateTime]
RegularAxisTuple = typing.Tuple[RealNumber, RealNumber, int]
GeneralAxisVector = np.ndarray


class Axis:
    """Axis class"""

    def __init__(self, relative_axis: GeneralAxisVector, origin: AxisStartType = 0):
        """
        Constructs an axis in the form origin + relative_axis that starts at origin + relative_axis[0]

        Relative axis should be a monotone array either increasing or decreasing

        :param relative_axis: a numpy array of real numbers
        :param origin: either a real number or a  PreciseDataTime [optional, default 0]
        """
        valid, error_message = _validate_general_axis(relative_axis)
        if not valid:
            raise ValueError(error_message)

        self._relative_axis = relative_axis

        if not isinstance(origin, (PreciseDateTime, int, float)):
            raise ValueError('origin should be either a PreciseDateTime or a real number')

        self._origin = origin

        self._increasing = _is_increasing(self._relative_axis)

    @property
    def start(self) -> AxisStartType:
        """
        :return: origin + relative_axis[0]

        see also origin
        """
        return self._origin + self._relative_axis[0]

    @property
    def origin(self) -> AxisStartType:
        """
        :return: origin

        see also start
        """
        return self._origin

    @property
    def increasing(self) -> bool:
        """
        :return: whether the axis is monotone increasing
        """
        return self._increasing

    @property
    def decreasing(self) -> bool:
        """
        :return: whether the axis is monotone decreasing
        """
        return not self.increasing

    @property
    def mean_step(self) -> RealNumber:
        """
        :return: the average step
        """
        return np.diff(self._relative_axis).mean()

    @property
    def size(self) -> int:
        """Number of axis points"""
        return self._relative_axis.size

    @property
    def length(self) -> RealNumber:
        """
        :return length of the internal part of the axis (N - 1) * step
        """
        return abs(self._relative_axis[-1] - self._relative_axis[0])

    def get_array(self, start: int = 0, stop: int = None) -> np.ndarray:
        """
        Get the underlying axis (default) or a portion of it ([start, stop)) in the form of a numpy.ndarray.

        :param start: start index (included)
        :param stop: stop index (excluded)
        :return: the array
        """
        return self.get_relative_array(start, stop) + self._origin

    def get_relative_array(self, start=0, stop=None) -> np.ndarray:
        """
        Get the underlying relative axis (default) or a portion of it ([start, stop)) in the form of a numpy.ndarray.

        The result is relative to the origin, i.e. the axis = relative_array + origin.

        :param start: start index (included)
        :param stop: stop index (excluded)
        :return: an array of floats
        """
        stop = self.size if stop is None else stop
        _check_range(start, stop, self.size)
        return self._relative_axis[start: stop]

    def get_interval_id(self, values) -> np.ndarray:
        """
        Find the interval of the axis in which each values lies.

        If values lie outside the axis the closest intervals are returned.

        :param values: values to locate on the axis
        :return: np.ndarray with the position of the values in the axis
        """
        return _get_interval_id_not_regular_real_axis(self._relative_axis, self.increasing, values - self._origin)

    def get_interval_id_from_relative(self, values) -> np.ndarray:
        """
        Find the interval of the axis in which each values lies.
        Here values are relative to origin.

        :param values: values to locate on the axis
        :return: np.ndarray with the position of the values in the axis
        """
        return self.get_interval_id(values + self._origin)

    def interpolate(self, position):
        """
        For a given position it returns the value of the axis via linear interpolation.

        It coincides with accessing the array at the given position when the position is an integer,
        otherwise it interpolates the closest values

        :param position: the position (integer of float)
        :return: the corresponding axis point
        """
        return scipy.interpolate.interp1d(range(self.size), self._relative_axis)(position) + self._origin

    def __repr__(self):
        """
        :return: string representation of the object
        """
        axis_repr = "{classname} -- direction: {direction}:" + \
                    "{newline} start: {start}, end: {end}" + \
                    "{newline} size: {size}, length: {length}"
        return axis_repr.format(start=self.start, size=self.size,
                                end=self.start + self.length if self.increasing else -self.length,
                                newline=os.linesep + '--', classname=self.__class__.__name__,
                                direction="increasing" if self.increasing else "decreasing", length=self.length)


class RegularAxis(Axis):
    """Regular axis class"""

    def __init__(self, relative_axis: RegularAxisTuple, origin: AxisStartType = 0):
        """
        A RegularAxis is an Axis where the relative_axis is given in the form:
            (start [float, int], step[float, int], size [int)

        Step can be negative

        :param relative_axis: a tuple (start [float, int], step[float, int], size [int)
        :param origin: either a real number or a PreciseDataTime [optional, default 0]
        """
        # input validation
        valid, error_message = _validate_uniform_axis(relative_axis)
        if not valid:
            raise ValueError(error_message)

        # relative axis initialization
        start, step, size = relative_axis
        super(RegularAxis, self).__init__(np.asarray([start + step * k for k in range(size)]), origin)

    @property
    def step(self) -> RealNumber:
        """Axis step"""
        return self.mean_step

    def interpolate(self, position):
        """
        For a given position it returns the value of the axis via linear interpolation.

        out = position * step + start

        :param position: the position (integer of float)
        :return: the corresponding axis point
        """
        return position * self.step + self.start

    def get_interval_id(self, values) -> np.ndarray:
        """
        Find the interval of the axis in which each values lies.

        If values lie outside the axis the closest intervals are returned.

        :param values: values to locate on the axis
        :return: np.ndarray with the position of the values in the axis
        """
        return _get_interval_id_regular_real_axis(self._relative_axis[0], self.step, self.size, values - self._origin)


def _check_range(start, stop, size):
    if stop > size:
        raise ValueError("Stop ({stop}) should be smaller equal than size: ({size})".format(stop=stop, size=size))
    if start < 0:
        raise ValueError("Start ({start}) should be positive".format(start=start))


def _is_increasing(axis):
    steps = np.diff(axis)
    if (steps > 0).all():
        return True
    elif (steps < 0).all():
        return False
    else:
        raise RuntimeError("Expecting monotone axis")


def _get_interval_id_regular_real_axis(start: RealNumber, step: RealNumber, size: int, values) -> np.ndarray:
    """
    The interval id is intended as the interval starting at the returned position, containing the point.
    If the point is exactly an edge of the interval, the interval starting at the point should be returned.

    :param start: the start of the axis
    :param step: the step of the axis
    :param size: the size of the axis
    :param values: the values to be located in the array
    :return: np.ndarray of the same size as values
    """

    def to_int_and_clip(value):
        return np.clip(int(np.floor(value)), 0, size - 1)

    val = (values - start) / step

    if isinstance(val, np.ndarray):
        return np.array([to_int_and_clip(v) for v in val])
    else:
        return np.array([to_int_and_clip(val)])


def _get_interval_id_not_regular_real_axis(array: np.ndarray, increasing: bool, values: np.ndarray) -> np.ndarray:
    """
    The interval id is intended as the interval starting at the returned position, containing the point.
    If the point is exactly an edge of the interval, the interval starting at the point should be returned.

    :param array
    :param values: the values to be located in the array
    :return: np.ndarray of the same size as values
    """

    values = _utils.input_data_to_numpy_array_with_checks(values, ndim=1)
    out = np.empty((values.size,))
    for k, v in enumerate(values):
        closest_pos = np.argmin(np.abs((array - v)))
        if increasing:
            if array[closest_pos] > v:
                out[k] = closest_pos - 1
            else:
                out[k] = closest_pos
        else:
            if array[closest_pos] < v:
                out[k] = closest_pos - 1
            else:
                out[k] = closest_pos
    return np.clip(out.astype(int), 0, array.size - 1)


def _validate_uniform_axis(uniform_axis: tuple) -> typing.Tuple[bool, str]:
    if not isinstance(uniform_axis, tuple):
        return False, "relative axis should be a tuple"

    if not len(uniform_axis) == 3:  # start, step and size
        return False, "relative axis should be a tuple of size three (start, step, size)"

    if not isinstance(uniform_axis[0], (float, int)):  # start
        return False, "start of relative axis {} should be a real number".format(type(uniform_axis[0]))

    if not isinstance(uniform_axis[1], (float, int)):  # step
        return False, "Step of relative axis {} should be a real number".format(type(uniform_axis[1]))

    if not isinstance(uniform_axis[2], int):  # step
        return False, "Type of relative axis {} should be an integer".format(type(uniform_axis[2]))

    if uniform_axis[2] < 1:  # step
        return False, "Size of relative axis {} should be an integer".format(uniform_axis[2])

    return True, ""


def _validate_general_axis(general_axis: np.ndarray) -> typing.Tuple[bool, str]:
    if not isinstance(general_axis, np.ndarray):
        return False, "relative axis should be a numpy.ndarray"

    if not ((general_axis.dtype == int) or (general_axis.dtype == float)):
        return False, "relative axis dtype {} should be a numpy.ndarray of real numbers".format(general_axis.dtype)

    if not general_axis.size == general_axis.shape[0]:
        return False, 'relative axis should be a vector not a matrix {} != {}'.format(general_axis.size,
                                                                                      general_axis.shape[0])

    return True, ""
